@@ -2,7 +2,8 @@
 #
 # ContentTypes - A class for writing the Excel XLSX ContentTypes file.
 #
-# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 import copy
@@ -160,6 +161,11 @@ class ContentTypes(xmlwriter.XMLwriter):
         self._add_override(('/docProps/custom.xml',
                            app_document + 'custom-properties+xml'))
 
+    def _add_metadata(self):
+        # Add the metadata file to the ContentTypes overrides.
+        self._add_override(('/xl/metadata.xml',
+                           app_document + 'spreadsheetml.sheetMetadata+xml'))
+
     ###########################################################################
     #
     # XML methods.
