@@ -2,7 +2,8 @@
 #
 # Drawing - A class for writing the Excel XLSX Drawing file.
 #
-# Copyright 2013-2019, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 from . import xmlwriter
@@ -51,14 +52,11 @@ class Drawing(xmlwriter.XMLwriter):
         self._write_drawing_workspace()
 
         if self.embedded:
-            index = 1
-            for drawing in self.drawings:
+            index = 0
+            for drawing_properties in self.drawings:
                 # Write the xdr:twoCellAnchor element.
-                self._write_two_cell_anchor(index, drawing)
                 index += 1
-
-                if drawing['url']:
-                    index += 1
+                self._write_two_cell_anchor(index, drawing_properties)
 
         else:
             # Write the xdr:absoluteAnchor element.
@@ -69,36 +67,27 @@ class Drawing(xmlwriter.XMLwriter):
         # Close the file.
         self._xml_close()
 
-    def _add_drawing_object(self, drawing_object):
+    def _add_drawing_object(self):
         # Add a chart, image or shape sub object to the drawing.
 
-        obj = {
-            'anchor_type': drawing_object[0],
-            'col_from': drawing_object[1],
-            'row_from': drawing_object[2],
-            'col_from_offset': drawing_object[3],
-            'row_from_offset': drawing_object[4],
-            'col_to': drawing_object[5],
-            'row_to': drawing_object[6],
-            'col_to_offset': drawing_object[7],
-            'row_to_offset': drawing_object[8],
-            'col_absolute': drawing_object[9],
-            'row_absolute': drawing_object[10],
-            'width': drawing_object[11],
-            'height': drawing_object[12],
-            'description': drawing_object[13],
-            'shape': drawing_object[14],
-            'url': None,
+        drawing_object = {
+            'anchor_type': None,
+            'dimensions': [],
+            'width': 0,
+            'height': 0,
+            'shape': None,
+            'anchor': None,
+            'rel_index': 0,
+            'url_rel_index': 0,
             'tip': None,
-            'anchor': None
+            'name': None,
+            'description': None,
+            'decorative': False
         }
 
-        if len(drawing_object) > 15:
-            obj['url'] = drawing_object[15]
-            obj['tip'] = drawing_object[16]
-            obj['anchor'] = drawing_object[17]
+        self.drawings.append(drawing_object)
 
-        self.drawings.append(obj)
+        return drawing_object
 
     ###########################################################################
     #
@@ -119,22 +108,37 @@ class Drawing(xmlwriter.XMLwriter):
 
         self._xml_start_tag('xdr:wsDr', attributes)
 
-    def _write_two_cell_anchor(self, index, drawing):
+    def _write_two_cell_anchor(self, index, drawing_properties):
         # Write the <xdr:twoCellAnchor> element.
-        shape = drawing['shape']
-
-        options = {
-            'description': drawing['description'],
-            'url': drawing['url'],
-            'tip': drawing['tip']
-        }
+        anchor_type = drawing_properties['type']
+        dimensions = drawing_properties['dimensions']
+        col_from = dimensions[0]
+        row_from = dimensions[1]
+        col_from_offset = dimensions[2]
+        row_from_offset = dimensions[3]
+        col_to = dimensions[4]
+        row_to = dimensions[5]
+        col_to_offset = dimensions[6]
+        row_to_offset = dimensions[7]
+        col_absolute = dimensions[8]
+        row_absolute = dimensions[9]
+        width = drawing_properties['width']
+        height = drawing_properties['height']
+        shape = drawing_properties['shape']
+        anchor = drawing_properties['anchor']
+        rel_index = drawing_properties['rel_index']
+        url_rel_index = drawing_properties['url_rel_index']
+        tip = drawing_properties['tip']
+        name = drawing_properties['name']
+        description = drawing_properties['description']
+        decorative = drawing_properties['decorative']
 
         attributes = []
 
         # Add attribute for positioning.
-        if drawing['anchor'] == 2:
+        if anchor == 2:
             attributes.append(('editAs', 'oneCell'))
-        elif drawing['anchor'] == 3:
+        elif anchor == 3:
             attributes.append(('editAs', 'absolute'))
 
         # Add editAs attribute for shapes.
@@ -145,39 +149,48 @@ class Drawing(xmlwriter.XMLwriter):
 
         # Write the xdr:from element.
         self._write_from(
-            drawing['col_from'],
-            drawing['row_from'],
-            drawing['col_from_offset'],
-            drawing['row_from_offset'])
+            col_from,
+            row_from,
+            col_from_offset,
+            row_from_offset)
 
         # Write the xdr:from element.
         self._write_to(
-            drawing['col_to'],
-            drawing['row_to'],
-            drawing['col_to_offset'],
-            drawing['row_to_offset'])
+            col_to,
+            row_to,
+            col_to_offset,
+            row_to_offset)
 
-        if drawing['anchor_type'] == 1:
+        if anchor_type == 1:
             # Graphic frame.
             # Write the xdr:graphicFrame element for charts.
-            self._write_graphic_frame(index, drawing['description'])
-        elif drawing['anchor_type'] == 2:
+            self._write_graphic_frame(index, rel_index, name,
+                                      description, decorative)
+        elif anchor_type == 2:
             # Write the xdr:pic element.
             self._write_pic(index,
-                            drawing['col_absolute'],
-                            drawing['row_absolute'],
-                            drawing['width'],
-                            drawing['height'],
+                            rel_index,
+                            col_absolute,
+                            row_absolute,
+                            width,
+                            height,
                             shape,
-                            options)
+                            description,
+                            url_rel_index,
+                            tip,
+                            decorative)
         else:
             # Write the xdr:sp element for shapes.
             self._write_sp(index,
-                           drawing['col_absolute'],
-                           drawing['row_absolute'],
-                           drawing['width'],
-                           drawing['height'],
-                           shape)
+                           col_absolute,
+                           row_absolute,
+                           width,
+                           height,
+                           shape,
+                           description,
+                           url_rel_index,
+                           tip,
+                           decorative)
 
         # Write the xdr:clientData element.
         self._write_client_data()
@@ -188,23 +201,23 @@ class Drawing(xmlwriter.XMLwriter):
         self._xml_start_tag('xdr:absoluteAnchor')
         # Write the <xdr:absoluteAnchor> element.
 
-        # Different co-ordinates for horizontal (= 0) and vertical (= 1).
+        # Different coordinates for horizontal (= 0) and vertical (= 1).
         if self.orientation == 0:
             # Write the xdr:pos element.
             self._write_pos(0, 0)
 
             # Write the xdr:ext element.
-            self._write_ext(9308969, 6078325)
+            self._write_xdr_ext(9308969, 6078325)
 
         else:
             # Write the xdr:pos element.
             self._write_pos(0, -47625)
 
             # Write the xdr:ext element.
-            self._write_ext(6162675, 6124575)
+            self._write_xdr_ext(6162675, 6124575)
 
         # Write the xdr:graphicFrame element.
-        self._write_graphic_frame(frame_index)
+        self._write_graphic_frame(frame_index, frame_index)
 
         # Write the xdr:clientData element.
         self._write_client_data()
@@ -270,31 +283,32 @@ class Drawing(xmlwriter.XMLwriter):
 
         self._xml_empty_tag('xdr:pos', attributes)
 
-    def _write_ext(self, cx, cy):
+    def _write_xdr_ext(self, cx, cy):
         # Write the <xdr:ext> element.
 
         attributes = [('cx', cx), ('cy', cy)]
 
         self._xml_empty_tag('xdr:ext', attributes)
 
-    def _write_graphic_frame(self, index, name=None):
+    def _write_graphic_frame(self, index, rel_index, name=None,
+                             description=None, decorative=None):
         # Write the <xdr:graphicFrame> element.
         attributes = [('macro', '')]
 
         self._xml_start_tag('xdr:graphicFrame', attributes)
 
         # Write the xdr:nvGraphicFramePr element.
-        self._write_nv_graphic_frame_pr(index, name)
+        self._write_nv_graphic_frame_pr(index, name, description, decorative)
 
         # Write the xdr:xfrm element.
         self._write_xfrm()
 
         # Write the a:graphic element.
-        self._write_atag_graphic(index)
+        self._write_atag_graphic(rel_index)
 
         self._xml_end_tag('xdr:graphicFrame')
 
-    def _write_nv_graphic_frame_pr(self, index, name):
+    def _write_nv_graphic_frame_pr(self, index, name, description, decorative):
         # Write the <xdr:nvGraphicFramePr> element.
 
         if not name:
@@ -303,44 +317,96 @@ class Drawing(xmlwriter.XMLwriter):
         self._xml_start_tag('xdr:nvGraphicFramePr')
 
         # Write the xdr:cNvPr element.
-        self._write_c_nv_pr(index + 1, name)
+        self._write_c_nv_pr(index + 1, name, description,
+                            None, None, decorative)
 
         # Write the xdr:cNvGraphicFramePr element.
         self._write_c_nv_graphic_frame_pr()
 
         self._xml_end_tag('xdr:nvGraphicFramePr')
 
-    def _write_c_nv_pr(self, index, name, options=None):
+    def _write_c_nv_pr(self, index, name, description, url_rel_index,
+                       tip, decorative):
         # Write the <xdr:cNvPr> element.
-        if options is None:
-            options = {}
-
-        descr = options.get('description', None)
-        url = options.get('url', None)
-        tip = options.get('tip', None)
-
         attributes = [('id', index), ('name', name)]
 
         # Add description attribute for images.
-        if descr is not None:
-            attributes.append(('descr', descr))
+        if description and not decorative:
+            attributes.append(('descr', description))
 
-        if url:
+        if url_rel_index or decorative:
             self._xml_start_tag('xdr:cNvPr', attributes)
-            schema = "http://schemas.openxmlformats.org"
-            att = [
-                ('xmlns:r', schema + "/officeDocument/2006/relationships"),
-                ('r:id', "rId" + str(index - 1))
-            ]
 
-            if tip:
-                att.append(('tooltip', tip))
+            if url_rel_index:
+                self._write_a_hlink_click(url_rel_index, tip)
+
+            if decorative:
+                self._write_decorative()
 
-            self._xml_empty_tag('a:hlinkClick', att)
             self._xml_end_tag('xdr:cNvPr')
         else:
             self._xml_empty_tag('xdr:cNvPr', attributes)
 
+    def _write_decorative(self):
+
+        self._xml_start_tag('a:extLst')
+
+        self._write_uri_ext('{FF2B5EF4-FFF2-40B4-BE49-F238E27FC236}')
+        self._write_a16_creation_id()
+        self._xml_end_tag('a:ext')
+
+        self._write_uri_ext('{C183D7F6-B498-43B3-948B-1728B52AA6E4}')
+        self._write_adec_decorative()
+        self._xml_end_tag('a:ext')
+
+        self._xml_end_tag('a:extLst')
+
+    def _write_uri_ext(self, uri):
+        # Write the <a:ext> element.
+        attributes = [('uri', uri)]
+
+        self._xml_start_tag('a:ext', attributes)
+
+    def _write_adec_decorative(self):
+        # Write the <adec:decorative> element.
+        xmlns = 'http://schemas.microsoft.com/office/drawing/2017/decorative'
+        val = '1'
+
+        attributes = [
+            ('xmlns:adec', xmlns),
+            ('val', val),
+        ]
+
+        self._xml_empty_tag('adec:decorative', attributes)
+
+    def _write_a16_creation_id(self):
+        # Write the <a16:creationId> element.
+
+        xmlns_a_16 = 'http://schemas.microsoft.com/office/drawing/2014/main'
+        id = '{00000000-0008-0000-0000-000002000000}'
+
+        attributes = [
+            ('xmlns:a16', xmlns_a_16),
+            ('id', id),
+        ]
+
+        self._xml_empty_tag('a16:creationId', attributes)
+
+    def _write_a_hlink_click(self, rel_index, tip):
+        # Write the <a:hlinkClick> element.
+        schema = 'http://schemas.openxmlformats.org/officeDocument/'
+        xmlns_r = schema + '2006/relationships'
+
+        attributes = [
+            ('xmlns:r', xmlns_r),
+            ('r:id', 'rId' + str(rel_index)),
+        ]
+
+        if tip:
+            attributes.append(('tooltip', tip))
+
+        self._xml_empty_tag('a:hlinkClick', attributes)
+
     def _write_c_nv_graphic_frame_pr(self):
         # Write the <xdr:cNvGraphicFramePr> element.
         if self.embedded:
@@ -433,7 +499,8 @@ class Drawing(xmlwriter.XMLwriter):
         self._xml_empty_tag('xdr:clientData')
 
     def _write_sp(self, index, col_absolute, row_absolute,
-                  width, height, shape):
+                  width, height, shape, description,
+                  url_rel_index, tip, decorative):
         # Write the <xdr:sp> element.
 
         if shape and shape.connect:
@@ -451,12 +518,13 @@ class Drawing(xmlwriter.XMLwriter):
         else:
             # Add attribute for shapes.
             attributes = [('macro', ''),
-                          ('textlink', '')]
+                          ('textlink', shape.textlink)]
 
             self._xml_start_tag('xdr:sp', attributes)
 
             # Write the xdr:nvSpPr element.
-            self._write_nv_sp_pr(index, shape)
+            self._write_nv_sp_pr(index, shape, url_rel_index, tip,
+                                 description, decorative)
 
             # Write the xdr:spPr element.
             self._write_xdr_sp_pr(index, col_absolute, row_absolute, width,
@@ -478,7 +546,7 @@ class Drawing(xmlwriter.XMLwriter):
 
         name = shape.name + ' ' + str(index)
         if name is not None:
-            self._write_c_nv_pr(index, name)
+            self._write_c_nv_pr(index, name, None, None, None, None)
 
         self._xml_start_tag('xdr:cNvCxnSpPr')
 
@@ -496,7 +564,8 @@ class Drawing(xmlwriter.XMLwriter):
         self._xml_end_tag('xdr:cNvCxnSpPr')
         self._xml_end_tag('xdr:nvCxnSpPr')
 
-    def _write_nv_sp_pr(self, index, shape):
+    def _write_nv_sp_pr(self, index, shape, url_rel_index, tip,
+                        description, decorative):
         # Write the <xdr:NvSpPr> element.
         attributes = []
 
@@ -504,32 +573,27 @@ class Drawing(xmlwriter.XMLwriter):
 
         name = shape.name + ' ' + str(index)
 
-        self._write_c_nv_pr(index + 1, name)
+        self._write_c_nv_pr(index + 1, name, description,
+                            url_rel_index, tip, decorative)
 
         if shape.name == 'TextBox':
             attributes = [('txBox', 1)]
 
         self._xml_empty_tag('xdr:cNvSpPr', attributes)
 
-        # attributes = [('noChangeArrowheads', '1')]
-        # self._xml_empty_tag('a:spLocks', attributes)
-        # self._xml_end_tag('xdr:cNvSpPr')
-
         self._xml_end_tag('xdr:nvSpPr')
 
-    def _write_pic(self, index, col_absolute, row_absolute,
-                   width, height, shape, options):
+    def _write_pic(self, index, rel_index, col_absolute, row_absolute,
+                   width, height, shape, description, url_rel_index,
+                   tip, decorative):
         # Write the <xdr:pic> element.
         self._xml_start_tag('xdr:pic')
 
         # Write the xdr:nvPicPr element.
-        self._write_nv_pic_pr(index, options)
-
+        self._write_nv_pic_pr(index, rel_index, description,
+                              url_rel_index, tip, decorative)
         # Write the xdr:blipFill element.
-        if options.get('url', None):
-            index = index + 1
-
-        self._write_blip_fill(index)
+        self._write_blip_fill(rel_index)
 
         # Write the xdr:spPr element.
         self._write_sp_pr(col_absolute, row_absolute, width, height,
@@ -537,12 +601,14 @@ class Drawing(xmlwriter.XMLwriter):
 
         self._xml_end_tag('xdr:pic')
 
-    def _write_nv_pic_pr(self, index, options):
+    def _write_nv_pic_pr(self, index, rel_index, description,
+                         url_rel_index, tip, decorative):
         # Write the <xdr:nvPicPr> element.
         self._xml_start_tag('xdr:nvPicPr')
 
         # Write the xdr:cNvPr element.
-        self._write_c_nv_pr(index + 1, 'Picture ' + str(index), options)
+        self._write_c_nv_pr(index + 1, 'Picture ' + str(index), description,
+                            url_rel_index, tip, decorative)
 
         # Write the xdr:cNvPicPr element.
         self._write_c_nv_pic_pr()
@@ -802,10 +868,20 @@ class Drawing(xmlwriter.XMLwriter):
 
     def _write_tx_body(self, col_absolute, row_absolute, width, height, shape):
         # Write the <xdr:txBody> element.
-        attributes = [
-            ('wrap', "square"),
-            ('rtlCol', "0"),
-        ]
+        attributes = []
+
+        if shape.text_rotation != 0:
+            if shape.text_rotation == 90:
+                attributes.append(('vert', 'vert270'))
+            if shape.text_rotation == -90:
+                attributes.append(('vert', 'vert'))
+            if shape.text_rotation == 270:
+                attributes.append(('vert', 'wordArtVert'))
+            if shape.text_rotation == 271:
+                attributes.append(('vert', 'eaVert'))
+
+        attributes.append(('wrap', 'square'))
+        attributes.append(('rtlCol', "0"))
 
         if not shape.align['defined']:
             attributes.append(('anchor', 't'))
@@ -841,23 +917,48 @@ class Drawing(xmlwriter.XMLwriter):
         latin_attrs = Shape._get_font_latin_attributes(font)
         style_attrs.insert(0, ('lang', font['lang']))
 
-        for line in lines:
+        if shape.textlink != '':
+            attributes = [
+                ('id', '{B8ADDEFE-BF52-4FD4-8C5D-6B85EF6FF707}'),
+                ('type', 'TxLink')]
+
             self._xml_start_tag('a:p')
-
-            if line == '':
-                self._write_font_run(font, style_attrs, latin_attrs,
-                                     'a:endParaRPr')
-                self._xml_end_tag('a:p')
-                continue
-
-            self._xml_start_tag('a:r')
+            self._xml_start_tag('a:fld', attributes)
 
             self._write_font_run(font, style_attrs, latin_attrs, 'a:rPr')
 
-            self._xml_data_element('a:t', line)
+            self._xml_data_element('a:t', shape.text)
+            self._xml_end_tag('a:fld')
+
+            self._write_font_run(font, style_attrs, latin_attrs,
+                                 'a:endParaRPr')
 
-            self._xml_end_tag('a:r')
             self._xml_end_tag('a:p')
+        else:
+            for line in lines:
+                self._xml_start_tag('a:p')
+
+                if line == '':
+                    self._write_font_run(font, style_attrs, latin_attrs,
+                                         'a:endParaRPr')
+                    self._xml_end_tag('a:p')
+                    continue
+                elif 'text' in shape.align:
+                    if shape.align['text'] == 'left':
+                        self._xml_empty_tag('a:pPr', [('algn', 'l')])
+                    if shape.align['text'] == 'center':
+                        self._xml_empty_tag('a:pPr', [('algn', 'ctr')])
+                    if shape.align['text'] == 'right':
+                        self._xml_empty_tag('a:pPr', [('algn', 'r')])
+
+                self._xml_start_tag('a:r')
+
+                self._write_font_run(font, style_attrs, latin_attrs, 'a:rPr')
+
+                self._xml_data_element('a:t', line)
+
+                self._xml_end_tag('a:r')
+                self._xml_end_tag('a:p')
 
         self._xml_end_tag('xdr:txBody')
 
