@@ -2,7 +2,8 @@
 #
 # Exceptions - A class for XlsxWriter exceptions.
 #
-# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 
@@ -34,6 +35,10 @@ class DuplicateWorksheetName(XlsxInputError):
     """Worksheet name already exists."""
 
 
+class OverlappingRange(XlsxInputError):
+    """Worksheet merge range or table overlaps previous range."""
+
+
 class UndefinedImageSize(XlsxFileError):
     """No size data found in image file."""
 
