@@ -2,11 +2,13 @@
 #
 # Format - A class for writing the Excel XLSX Worksheet file.
 #
-# Copyright 2013-2017, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 # Package imports.
 from . import xmlwriter
+from warnings import warn
 
 
 class Format(xmlwriter.XMLwriter):
@@ -37,7 +39,7 @@ class Format(xmlwriter.XMLwriter):
         self.xf_index = None
         self.dxf_index = None
 
-        self.num_format = 0
+        self.num_format = 'General'
         self.num_format_index = 0
         self.font_index = 0
         self.has_font = 0
@@ -104,10 +106,14 @@ class Format(xmlwriter.XMLwriter):
         self.color_indexed = 0
         self.font_only = 0
 
+        self.quote_prefix = False
+
         # Convert properties in the constructor to method calls.
         for key, value in properties.items():
             getattr(self, 'set_' + key)(value)
 
+        self._format_key = None
+
     ###########################################################################
     #
     # Format properties.
@@ -378,8 +384,8 @@ class Format(xmlwriter.XMLwriter):
             if rotation < 0:
                 rotation = -rotation + 90
         else:
-            raise Exception(
-                "Rotation rotation outside range: -90 <= angle <= 90")
+            warn("Rotation rotation outside range: -90 <= angle <= 90")
+            return
 
         self.rotation = rotation
 
@@ -638,6 +644,19 @@ class Format(xmlwriter.XMLwriter):
         """
         self.diag_border = diag_border
 
+    def set_quote_prefix(self, quote_prefix=True):
+        """
+        Set the Format quote prefix property.
+
+        Args:
+            quote_prefix: Default is True, turns property on.
+
+        Returns:
+            Nothing.
+
+        """
+        self.quote_prefix = quote_prefix
+
     ###########################################################################
     #
     # Internal Format properties. These aren't documented since they are
@@ -677,9 +696,9 @@ class Format(xmlwriter.XMLwriter):
         # Set the text_v_align property.
         self.text_v_align = text_v_align
 
-    def set_reading_order(self, reading_order=True):
+    def set_reading_order(self, direction=0):
         # Set the reading_order property.
-        self.reading_order = reading_order
+        self.reading_order = direction
 
     def set_valign(self, align):
         # Set vertical cell alignment. This is required by the constructor
@@ -842,21 +861,23 @@ class Format(xmlwriter.XMLwriter):
         return attribs
 
     def _get_format_key(self):
-        # Returns a unique hash key for a font. Used by Workbook.
-        key = ':'.join(self._to_string(x) for x in (
-            self._get_font_key(),
-            self._get_border_key(),
-            self._get_fill_key(),
-            self._get_alignment_key(),
-            self.num_format,
-            self.locked,
-            self.hidden))
+        # Returns a unique hash key for a format. Used by Workbook.
+        if self._format_key is None:
+            self._format_key = ':'.join(str(x) for x in (
+                self._get_font_key(),
+                self._get_border_key(),
+                self._get_fill_key(),
+                self._get_alignment_key(),
+                self.num_format,
+                self.locked,
+                self.quote_prefix,
+                self.hidden))
 
-        return key
+        return self._format_key
 
     def _get_font_key(self):
         # Returns a unique hash key for a font. Used by Workbook.
-        key = ':'.join(self._to_string(x) for x in (
+        key = ':'.join(str(x) for x in (
             self.bold,
             self.font_color,
             self.font_charset,
@@ -875,7 +896,7 @@ class Format(xmlwriter.XMLwriter):
 
     def _get_border_key(self):
         # Returns a unique hash key for a border style. Used by Workbook.
-        key = ':'.join(self._to_string(x) for x in (
+        key = ':'.join(str(x) for x in (
             self.bottom,
             self.bottom_color,
             self.diag_border,
@@ -892,7 +913,7 @@ class Format(xmlwriter.XMLwriter):
 
     def _get_fill_key(self):
         # Returns a unique hash key for a fill style. Used by Workbook.
-        key = ':'.join(self._to_string(x) for x in (
+        key = ':'.join(str(x) for x in (
             self.pattern,
             self.bg_color,
             self.fg_color))
@@ -902,7 +923,7 @@ class Format(xmlwriter.XMLwriter):
     def _get_alignment_key(self):
         # Returns a unique hash key for alignment formats.
 
-        key = ':'.join(self._to_string(x) for x in (
+        key = ':'.join(str(x) for x in (
             self.text_h_align,
             self.text_v_align,
             self.indent,
@@ -980,13 +1001,6 @@ class Format(xmlwriter.XMLwriter):
 
         return color
 
-    def _to_string(self, value):
-        # Convert number to a string but allow for utf-8 strings in Python 2.
-        try:
-            return str(value)
-        except UnicodeEncodeError:
-            return value.encode('utf-8')
-
     ###########################################################################
     #
     # XML methods.
