@@ -2,7 +2,8 @@
 #
 # Packager - A class for writing the Excel XLSX Worksheet file.
 #
-# Copyright 2013-2019, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 # Standard packages.
@@ -11,7 +12,7 @@ import stat
 import tempfile
 from shutil import copy
 
-from .compatibility import StringIO
+from io import StringIO
 from io import BytesIO
 
 # Package imports.
@@ -19,6 +20,7 @@ from .app import App
 from .contenttypes import ContentTypes
 from .core import Core
 from .custom import Custom
+from .metadata import Metadata
 from .relationships import Relationships
 from .sharedstrings import SharedStrings
 from .styles import Styles
@@ -129,6 +131,9 @@ class Packager(object):
 
     def _create_package(self):
         # Write the xml files that make up the XLSX OPC package.
+        self._write_content_types_file()
+        self._write_root_rels_file()
+        self._write_workbook_rels_file()
         self._write_worksheet_files()
         self._write_chartsheet_files()
         self._write_workbook_file()
@@ -138,19 +143,17 @@ class Packager(object):
         self._write_comment_files()
         self._write_table_files()
         self._write_shared_strings_file()
-        self._write_app_file()
-        self._write_core_file()
-        self._write_custom_file()
-        self._write_content_types_file()
         self._write_styles_file()
+        self._write_custom_file()
         self._write_theme_file()
-        self._write_root_rels_file()
-        self._write_workbook_rels_file()
         self._write_worksheet_rels_files()
         self._write_chartsheet_rels_files()
         self._write_drawing_rels_files()
         self._add_image_files()
         self._add_vba_project()
+        self._write_core_file()
+        self._write_app_file()
+        self._write_metadata_file()
 
         return self.filenames
 
@@ -318,6 +321,7 @@ class Packager(object):
             app._add_part_name(named_range)
 
         app._set_properties(properties)
+        app.doc_security = self.workbook.read_only
 
         app._set_xml_writer(self._filename('docProps/app.xml'))
         app._assemble_xml_file()
@@ -331,6 +335,15 @@ class Packager(object):
         core._set_xml_writer(self._filename('docProps/core.xml'))
         core._assemble_xml_file()
 
+    def _write_metadata_file(self):
+        # Write the metadata.xml file.
+        if not self.workbook.has_metadata:
+            return
+
+        metadata = Metadata()
+        metadata._set_xml_writer(self._filename('xl/metadata.xml'))
+        metadata._assemble_xml_file()
+
     def _write_custom_file(self):
         # Write the custom.xml file.
         properties = self.workbook.custom_properties
@@ -348,6 +361,8 @@ class Packager(object):
         content = ContentTypes()
         content._add_image_types(self.workbook.image_types)
 
+        self._get_table_count()
+
         worksheet_index = 1
         chartsheet_index = 1
         for worksheet in self.workbook.worksheets():
@@ -385,6 +400,10 @@ class Packager(object):
         if self.workbook.custom_properties:
             content._add_custom_properties()
 
+        # Add the metadata file if present.
+        if self.workbook.has_metadata:
+            content._add_metadata()
+
         content._set_xml_writer(self._filename('[Content_Types].xml'))
         content._assemble_xml_file()
 
@@ -393,22 +412,24 @@ class Packager(object):
         xf_formats = self.workbook.xf_formats
         palette = self.workbook.palette
         font_count = self.workbook.font_count
-        num_format_count = self.workbook.num_format_count
+        num_formats = self.workbook.num_formats
         border_count = self.workbook.border_count
         fill_count = self.workbook.fill_count
         custom_colors = self.workbook.custom_colors
         dxf_formats = self.workbook.dxf_formats
+        has_comments = self.workbook.has_comments
 
         styles = Styles()
         styles._set_style_properties([
             xf_formats,
             palette,
             font_count,
-            num_format_count,
+            num_formats,
             border_count,
             fill_count,
             custom_colors,
-            dxf_formats])
+            dxf_formats,
+            has_comments])
 
         styles._set_xml_writer(self._filename('xl/styles.xml'))
         styles._assemble_xml_file()
@@ -435,9 +456,14 @@ class Packager(object):
                                                      + str(index) + '.xml'))
                 table._set_properties(table_props)
                 table._assemble_xml_file()
-                self.table_count += 1
                 index += 1
 
+    def _get_table_count(self):
+        # Count the table files. Required for the [Content_Types] file.
+        for worksheet in self.workbook.worksheets():
+            for table_props in worksheet.tables:
+                self.table_count += 1
+
     def _write_root_rels_file(self):
         # Write the _rels/.rels xml file.
         rels = Relationships()
@@ -491,6 +517,10 @@ class Packager(object):
         if self.workbook.vba_project:
             rels._add_ms_package_relationship('/vbaProject', 'vbaProject.bin')
 
+        # Add the metadata file if required.
+        if self.workbook.has_metadata:
+            rels._add_document_relationship('/sheetMetadata', 'metadata.xml')
+
         rels._set_xml_writer(self._filename('xl/_rels/workbook.xml.rels'))
         rels._assemble_xml_file()
 
@@ -507,6 +537,7 @@ class Packager(object):
             external_links = (worksheet.external_hyper_links +
                               worksheet.external_drawing_links +
                               worksheet.external_vml_links +
+                              worksheet.external_background_links +
                               worksheet.external_table_links +
                               worksheet.external_comment_links)
 
