@@ -2,7 +2,8 @@
 #
 # Styles - A class for writing the Excel XLSX Worksheet file.
 #
-# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 # Package imports.
@@ -33,7 +34,7 @@ class Styles(xmlwriter.XMLwriter):
         self.xf_formats = []
         self.palette = []
         self.font_count = 0
-        self.num_format_count = 0
+        self.num_formats = []
         self.border_count = 0
         self.fill_count = 0
         self.custom_colors = []
@@ -99,7 +100,7 @@ class Styles(xmlwriter.XMLwriter):
         self.xf_formats = properties[0]
         self.palette = properties[1]
         self.font_count = properties[2]
-        self.num_format_count = properties[3]
+        self.num_formats = properties[3]
         self.border_count = properties[4]
         self.fill_count = properties[5]
         self.custom_colors = properties[6]
@@ -128,18 +129,15 @@ class Styles(xmlwriter.XMLwriter):
 
     def _write_num_fmts(self):
         # Write the <numFmts> element.
-        if not self.num_format_count:
+        if not self.num_formats:
             return
 
-        attributes = [('count', self.num_format_count)]
+        attributes = [('count', len(self.num_formats))]
         self._xml_start_tag('numFmts', attributes)
 
         # Write the numFmts elements.
-        for xf_format in self.xf_formats:
-            # Ignore built-in number formats, i.e., < 164.
-            if xf_format.num_format_index >= 164:
-                self._write_num_fmt(xf_format.num_format_index,
-                                    xf_format.num_format)
+        for index, num_format in enumerate(self.num_formats, 164):
+            self._write_num_fmt(index, num_format)
 
         self._xml_end_tag('numFmts')
 
@@ -387,6 +385,11 @@ class Styles(xmlwriter.XMLwriter):
             'gray0625',
         )
 
+        # Special handling for pattern only case.
+        if not fg_color and not bg_color and patterns[pattern]:
+            self._write_default_fill(patterns[pattern])
+            return
+
         self._xml_start_tag('fill')
 
         # The "none" pattern is handled differently for dxf formats.
@@ -405,7 +408,7 @@ class Styles(xmlwriter.XMLwriter):
             bg_color = self._get_palette_color(bg_color)
             self._xml_empty_tag('bgColor', [('rgb', bg_color)])
         else:
-            if not is_dxf_format:
+            if not is_dxf_format and pattern <= 1:
                 self._xml_empty_tag('bgColor', [('indexed', 64)])
 
         self._xml_end_tag('patternFill')
@@ -597,6 +600,9 @@ class Styles(xmlwriter.XMLwriter):
             ('xfId', xf_id),
         ]
 
+        if xf_format.quote_prefix:
+            attributes.append(('quotePrefix', 1))
+
         if xf_format.num_format_index > 0:
             attributes.append(('applyNumberFormat', 1))
 
