@@ -2,14 +2,34 @@
 #
 # Worksheet - A class for writing Excel Worksheets.
 #
-# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 import re
 import datetime
 from warnings import warn
 
 COL_NAMES = {}
-range_parts = re.compile(r'(\$?)([A-Z]{1,3})(\$?)(\d+)')
+
+CHAR_WIDTHS = {
+    " ": 3,  "!": 5,  '"': 6, "#": 7, "$": 7,  "%": 11, "&": 10, "'": 3,
+    "(": 5,  ")": 5,  "*": 7, "+": 7, ",": 4,  "-": 5,  ".": 4,  "/": 6,
+    "0": 7,  "1": 7,  "2": 7, "3": 7, "4": 7,  "5": 7,  "6": 7,  "7": 7,
+    "8": 7,  "9": 7,  ":": 4, ";": 4, "<": 7,  "=": 7,  ">": 7,  "?": 7,
+    "@": 13, "A": 9,  "B": 8, "C": 8, "D": 9,  "E": 7,  "F": 7,  "G": 9,
+    "H": 9,  "I": 4,  "J": 5, "K": 8, "L": 6,  "M": 12, "N": 10, "O": 10,
+    "P": 8,  "Q": 10, "R": 8, "S": 7, "T": 7,  "U": 9,  "V": 9,  "W": 13,
+    "X": 8,  "Y": 7,  "Z": 7, "[": 5, "\\": 6, "]": 5,  "^": 7,  "_": 7,
+    "`": 4,  "a": 7,  "b": 8, "c": 6, "d": 8,  "e": 8,  "f": 5,  "g": 7,
+    "h": 8,  "i": 4,  "j": 4, "k": 7, "l": 4,  "m": 12, "n": 8,  "o": 8,
+    "p": 8,  "q": 8,  "r": 5, "s": 6, "t": 5,  "u": 8,  "v": 7,  "w": 11,
+    "x": 7,  "y": 7,  "z": 6, "{": 5, "|": 7,  "}": 5,  "~": 7,
+}
+
+# Compile performance critical regular expressions.
+re_leading = re.compile(r'^\s')
+re_trailing = re.compile(r'\s$')
+re_range_parts = re.compile(r'(\$?)([A-Z]{1,3})(\$?)(\d+)')
 
 
 def xl_rowcol_to_cell(row, col, row_abs=False, col_abs=False):
@@ -117,7 +137,7 @@ def xl_cell_to_rowcol(cell_str):
     if not cell_str:
         return 0, 0
 
-    match = range_parts.match(cell_str)
+    match = re_range_parts.match(cell_str)
     col_str = match.group(2)
     row_str = match.group(4)
 
@@ -150,7 +170,7 @@ def xl_cell_to_rowcol_abs(cell_str):
     if not cell_str:
         return 0, 0, False, False
 
-    match = range_parts.match(cell_str)
+    match = re_range_parts.match(cell_str)
 
     col_abs = match.group(1)
     col_str = match.group(2)
@@ -202,7 +222,10 @@ def xl_range(first_row, first_col, last_row, last_col):
         warn("Row and column numbers must be >= 0")
         return None
 
-    return range1 + ':' + range2
+    if range1 == range2:
+        return range1
+    else:
+        return range1 + ':' + range2
 
 
 def xl_range_abs(first_row, first_col, last_row, last_col):
@@ -227,7 +250,10 @@ def xl_range_abs(first_row, first_col, last_row, last_col):
         warn("Row and column numbers must be >= 0")
         return None
 
-    return range1 + ':' + range2
+    if range1 == range2:
+        return range1
+    else:
+        return range1 + ':' + range2
 
 
 def xl_range_formula(sheetname, first_row, first_col, last_row, last_col):
@@ -270,12 +296,33 @@ def quote_sheetname(sheetname):
         # Double quote any single quotes.
         sheetname = sheetname.replace("'", "''")
 
-        # Singe quote the sheet name.
+        # Single quote the sheet name.
         sheetname = "'%s'" % sheetname
 
     return sheetname
 
 
+def xl_pixel_width(string):
+    """
+    Get the pixel width of a string based on individual character widths taken
+    from Excel. UTF8 characters, and other unhandled characters, are given a
+    default width of 8.
+
+    Args:
+       string: The string to calculate the width for. String.
+
+    Returns:
+        The string width in pixels. Note, Excel adds an additional 7 pixels of
+        padding in the cell.
+
+    """
+    length = 0
+    for char in string:
+        length += CHAR_WIDTHS.get(char, 8)
+
+    return length
+
+
 def xl_color(color):
     # Used in conjunction with the XlsxWriter *color() methods to convert
     # a color name into an RGB formatted string. These colors are for
@@ -624,10 +671,10 @@ def get_sparkline_style(style_id):
 
 def supported_datetime(dt):
     # Determine is an argument is a supported datetime object.
-    return(isinstance(dt, (datetime.datetime,
+    return isinstance(dt, (datetime.datetime,
                            datetime.date,
                            datetime.time,
-                           datetime.timedelta)))
+                           datetime.timedelta))
 
 
 def remove_datetime_timezone(dt_obj, remove_timezone):
@@ -651,6 +698,7 @@ def datetime_to_excel_datetime(dt_obj, date_1904, remove_timezone):
     # part of the number stores the number of days since the epoch and the
     # fractional part stores the percentage of the day.
     date_type = dt_obj
+    is_timedelta = False
 
     if date_1904:
         # Excel for Mac date epoch.
@@ -672,6 +720,7 @@ def datetime_to_excel_datetime(dt_obj, date_1904, remove_timezone):
         dt_obj = remove_datetime_timezone(dt_obj, remove_timezone)
         delta = dt_obj - epoch
     elif isinstance(dt_obj, datetime.timedelta):
+        is_timedelta = True
         delta = dt_obj
     else:
         raise TypeError("Unknown or unsupported datetime type")
@@ -690,7 +739,16 @@ def datetime_to_excel_datetime(dt_obj, date_1904, remove_timezone):
         excel_time -= 1
 
     # Account for Excel erroneously treating 1900 as a leap year.
-    if not date_1904 and excel_time > 59:
+    if not date_1904 and not is_timedelta and excel_time > 59:
         excel_time += 1
 
     return excel_time
+
+
+def preserve_whitespace(string):
+    # Check if a string has leading or trailing whitespace that requires a
+    # "preserve" attribute.
+    if (re_leading.search(string) or re_trailing.search(string)):
+        return True
+    else:
+        return False
