@@ -2,21 +2,23 @@
 #
 # Workbook - A class for writing the Excel XLSX Workbook file.
 #
-# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
+# SPDX-License-Identifier: BSD-2-Clause
+# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
 #
 
 # Standard packages.
 import hashlib
-import re
-import os
 import operator
+import os
+import re
 import time
-from warnings import warn
 from datetime import datetime
-from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED, LargeZipFile
+from decimal import Decimal
+from fractions import Fraction
 from struct import unpack
+from warnings import warn
+from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED, LargeZipFile
 
-from .compatibility import int_types, num_types, str_types, force_unicode
 
 # Package imports.
 from . import xmlwriter
@@ -81,6 +83,7 @@ class Workbook(xmlwriter.XMLwriter):
         self.in_memory = options.get('in_memory', False)
         self.excel2003_style = options.get('excel2003_style', False)
         self.remove_timezone = options.get('remove_timezone', False)
+        self.use_future_functions = options.get('use_future_functions', False)
         self.default_format_properties = \
             options.get('default_format_properties', {})
 
@@ -88,6 +91,11 @@ class Workbook(xmlwriter.XMLwriter):
         if self.max_url_length < 255:
             self.max_url_length = 2079
 
+        if options.get('use_zip64'):
+            self.allow_zip64 = True
+        else:
+            self.allow_zip64 = False
+
         self.worksheet_meta = WorksheetMeta()
         self.selected = 0
         self.fileclosed = 0
@@ -108,7 +116,7 @@ class Workbook(xmlwriter.XMLwriter):
         self.dxf_format_indices = {}
         self.palette = []
         self.font_count = 0
-        self.num_format_count = 0
+        self.num_formats = []
         self.defined_names = []
         self.named_ranges = []
         self.custom_colors = []
@@ -133,9 +141,10 @@ class Workbook(xmlwriter.XMLwriter):
         self.drawing_count = 0
         self.calc_mode = "auto"
         self.calc_on_load = True
-        self.allow_zip64 = False
         self.calc_id = 124519
         self.has_comments = False
+        self.read_only = 0
+        self.has_metadata = False
 
         # We can't do 'constant_memory' mode while doing 'in_memory' mode.
         if self.in_memory:
@@ -289,8 +298,7 @@ class Workbook(xmlwriter.XMLwriter):
 
         """
         if not is_stream and not os.path.exists(vba_project):
-            warn("VBA project binary file '%s' not found."
-                 % force_unicode(vba_project))
+            warn("VBA project binary file '%s' not found." % vba_project)
             return -1
 
         if self.vba_codename is None:
@@ -316,12 +324,20 @@ class Workbook(xmlwriter.XMLwriter):
                 self._store_workbook()
             except IOError as e:
                 raise FileCreateError(e)
-            except LargeZipFile as e:
+            except LargeZipFile:
                 raise FileSizeError("Filesize would require ZIP64 extensions. "
                                     "Use workbook.use_zip64().")
 
             self.fileclosed = True
 
+            # Ensure all constant_memory temp files are closed.
+            if self.constant_memory:
+                for worksheet in self.worksheets():
+                    worksheet._opt_close()
+
+        else:
+            warn("Calling close() on already closed file.")
+
     def set_size(self, width, height):
         """
         Set the size of a workbook window.
@@ -401,9 +417,9 @@ class Workbook(xmlwriter.XMLwriter):
                 property_type = 'bool'
             elif isinstance(value, datetime):
                 property_type = 'date'
-            elif isinstance(value, int_types):
+            elif isinstance(value, int):
                 property_type = 'number_int'
-            elif isinstance(value, num_types):
+            elif isinstance(value, (float, int, Decimal, Fraction)):
                 property_type = 'number'
             else:
                 property_type = 'text'
@@ -413,13 +429,11 @@ class Workbook(xmlwriter.XMLwriter):
 
         if property_type == 'text' and len(value) > 255:
             warn("Length of 'value' parameter exceeds Excel's limit of 255 "
-                 "characters in set_custom_property(): '%s'" %
-                 force_unicode(value))
+                 "characters in set_custom_property(): '%s'" % value)
 
         if len(name) > 255:
             warn("Length of 'name' parameter exceeds Excel's limit of 255 "
-                 "characters in set_custom_property(): '%s'" %
-                 force_unicode(name))
+                 "characters in set_custom_property(): '%s'" % name)
 
         self.custom_properties.append((name, value, property_type))
 
@@ -480,8 +494,7 @@ class Workbook(xmlwriter.XMLwriter):
 
             # Warn if the sheet index wasn't found.
             if sheet_index is None:
-                warn("Unknown sheet name '%s' in defined_name()"
-                     % force_unicode(sheetname))
+                warn("Unknown sheet name '%s' in defined_name()" % sheetname)
                 return -1
         else:
             # Use -1 to indicate global names.
@@ -490,21 +503,19 @@ class Workbook(xmlwriter.XMLwriter):
         # Warn if the defined name contains invalid chars as defined by Excel.
         if (not re.match(r'^[\w\\][\w\\.]*$', name, re.UNICODE)
                 or re.match(r'^\d', name)):
-            warn("Invalid Excel characters in defined_name(): '%s'"
-                 % force_unicode(name))
+            warn("Invalid Excel characters in defined_name(): '%s'" % name)
             return -1
 
         # Warn if the defined name looks like a cell name.
         if re.match(r'^[a-zA-Z][a-zA-Z]?[a-dA-D]?[0-9]+$', name):
-            warn("Name looks like a cell name in defined_name(): '%s'"
-                 % force_unicode(name))
+            warn("Name looks like a cell name in defined_name(): '%s'" % name)
             return -1
 
         # Warn if the name looks like a R1C1 cell reference.
         if (re.match(r'^[rcRC]$', name)
                 or re.match(r'^[rcRC]\d+[rcRC]\d+$', name)):
             warn("Invalid name '%s' like a RC cell ref in defined_name()"
-                 % force_unicode(name))
+                 % name)
             return -1
 
         self.defined_names.append([name, sheet_index, formula, False])
@@ -580,6 +591,19 @@ class Workbook(xmlwriter.XMLwriter):
         else:
             self.vba_codename = 'ThisWorkbook'
 
+    def read_only_recommended(self):
+        """
+        Set the Excel "Read-only recommended" option when saving a file.
+
+        Args:
+            None.
+
+        Returns:
+            Nothing.
+
+        """
+        self.read_only = 2
+
     ###########################################################################
     #
     # Private API.
@@ -601,6 +625,9 @@ class Workbook(xmlwriter.XMLwriter):
         # Write the fileVersion element.
         self._write_file_version()
 
+        # Write the fileSharing element.
+        self._write_file_sharing()
+
         # Write the workbookPr element.
         self._write_workbook_pr()
 
@@ -672,6 +699,9 @@ class Workbook(xmlwriter.XMLwriter):
         # Prepare the worksheet tables.
         self._prepare_tables()
 
+        # Prepare the metadata file links.
+        self._prepare_metadata()
+
         # Package the workbook.
         packager._add_workbook(self)
         packager._set_tmpdir(self.tmpdir)
@@ -746,6 +776,7 @@ class Workbook(xmlwriter.XMLwriter):
             'excel2003_style': self.excel2003_style,
             'remove_timezone': self.remove_timezone,
             'max_url_length': self.max_url_length,
+            'use_future_functions': self.use_future_functions,
         }
 
         worksheet._initialize(init_data)
@@ -894,7 +925,8 @@ class Workbook(xmlwriter.XMLwriter):
 
     def _prepare_num_formats(self):
         # User defined records in Excel start from index 0xA4.
-        num_formats = {}
+        unique_num_formats = {}
+        num_formats = []
         index = 164
         num_format_count = 0
 
@@ -902,7 +934,7 @@ class Workbook(xmlwriter.XMLwriter):
             num_format = xf_format.num_format
 
             # Check if num_format is an index to a built-in number format.
-            if not isinstance(num_format, str_types):
+            if not isinstance(num_format, str):
                 num_format = int(num_format)
 
                 # Number format '0' is indexed as 1 in Excel.
@@ -920,20 +952,21 @@ class Workbook(xmlwriter.XMLwriter):
                 xf_format.num_format_index = 0
                 continue
 
-            if num_format in num_formats:
+            if num_format in unique_num_formats:
                 # Number xf_format has already been used.
-                xf_format.num_format_index = num_formats[num_format]
+                xf_format.num_format_index = unique_num_formats[num_format]
             else:
                 # Add a new number xf_format.
-                num_formats[num_format] = index
+                unique_num_formats[num_format] = index
                 xf_format.num_format_index = index
                 index += 1
 
                 # Only increase font count for XF formats (not DXF formats).
                 if xf_format.xf_index:
+                    num_formats.append(num_format)
                     num_format_count += 1
 
-        self.num_format_count = num_format_count
+        self.num_formats = num_formats
 
     def _prepare_borders(self):
         # Iterate through the XF Format objects and give them an index to
@@ -1007,7 +1040,6 @@ class Workbook(xmlwriter.XMLwriter):
 
             if (xf_format.pattern <= 1 and xf_format.bg_color == 0
                     and xf_format.fg_color != 0):
-                xf_format.bg_color = 0
                 xf_format.pattern = 1
 
             key = xf_format._get_fill_key()
@@ -1101,6 +1133,7 @@ class Workbook(xmlwriter.XMLwriter):
         drawing_id = 0
         image_ids = {}
         header_image_ids = {}
+        background_ids = {}
 
         for sheet in self.worksheets():
             chart_count = len(sheet.charts)
@@ -1109,10 +1142,12 @@ class Workbook(xmlwriter.XMLwriter):
 
             header_image_count = len(sheet.header_images)
             footer_image_count = len(sheet.footer_images)
+            has_background = sheet.background_image
             has_drawing = False
 
             if not (chart_count or image_count or shape_count
-                    or header_image_count or footer_image_count):
+                    or header_image_count or footer_image_count
+                    or has_background):
                 continue
 
             # Don't increase the drawing_id header/footer images.
@@ -1120,23 +1155,45 @@ class Workbook(xmlwriter.XMLwriter):
                 drawing_id += 1
                 has_drawing = True
 
+            # Prepare the background images.
+            if sheet.background_image:
+                if sheet.background_bytes:
+                    filename = ''
+                    image_data = sheet.background_image
+                else:
+                    filename = sheet.background_image
+                    image_data = None
+
+                (image_type, width, height, name, x_dpi, y_dpi, digest) = \
+                    self._get_image_properties(filename, image_data)
+
+                if digest in background_ids:
+                    ref_id = background_ids[digest]
+                else:
+                    image_ref_id += 1
+                    ref_id = image_ref_id
+                    background_ids[digest] = image_ref_id
+                    self.images.append([filename, image_type, image_data])
+
+                sheet._prepare_background(ref_id, image_type)
+
             # Prepare the worksheet images.
             for index in range(image_count):
                 filename = sheet.images[index][2]
                 image_data = sheet.images[index][10]
-                (image_type, width, height, name, x_dpi, y_dpi, md5) = \
+                (image_type, width, height, name, x_dpi, y_dpi, digest) = \
                     self._get_image_properties(filename, image_data)
 
-                if md5 in image_ids:
-                    ref_id = image_ids[md5]
+                if digest in image_ids:
+                    ref_id = image_ids[digest]
                 else:
                     image_ref_id += 1
                     ref_id = image_ref_id
-                    image_ids[md5] = image_ref_id
+                    image_ids[digest] = image_ref_id
                     self.images.append([filename, image_type, image_data])
 
                 sheet._prepare_image(index, ref_id, drawing_id, width, height,
-                                     name, image_type, x_dpi, y_dpi, md5)
+                                     name, image_type, x_dpi, y_dpi, digest)
 
             # Prepare the worksheet charts.
             for index in range(chart_count):
@@ -1154,20 +1211,20 @@ class Workbook(xmlwriter.XMLwriter):
                 image_data = sheet.header_images[index][1]
                 position = sheet.header_images[index][2]
 
-                (image_type, width, height, name, x_dpi, y_dpi, md5) = \
+                (image_type, width, height, name, x_dpi, y_dpi, digest) = \
                     self._get_image_properties(filename, image_data)
 
-                if md5 in header_image_ids:
-                    ref_id = header_image_ids[md5]
+                if digest in header_image_ids:
+                    ref_id = header_image_ids[digest]
                 else:
                     image_ref_id += 1
                     ref_id = image_ref_id
-                    header_image_ids[md5] = image_ref_id
+                    header_image_ids[digest] = image_ref_id
                     self.images.append([filename, image_type, image_data])
 
                 sheet._prepare_header_image(ref_id, width, height,
                                             name, image_type, position,
-                                            x_dpi, y_dpi, md5)
+                                            x_dpi, y_dpi, digest)
 
             # Prepare the footer images.
             for index in range(footer_image_count):
@@ -1176,20 +1233,20 @@ class Workbook(xmlwriter.XMLwriter):
                 image_data = sheet.footer_images[index][1]
                 position = sheet.footer_images[index][2]
 
-                (image_type, width, height, name, x_dpi, y_dpi, md5) = \
+                (image_type, width, height, name, x_dpi, y_dpi, digest) = \
                     self._get_image_properties(filename, image_data)
 
-                if md5 in header_image_ids:
-                    ref_id = header_image_ids[md5]
+                if digest in header_image_ids:
+                    ref_id = header_image_ids[digest]
                 else:
                     image_ref_id += 1
                     ref_id = image_ref_id
-                    header_image_ids[md5] = image_ref_id
+                    header_image_ids[digest] = image_ref_id
                     self.images.append([filename, image_type, image_data])
 
                 sheet._prepare_header_image(image_ref_id, width, height,
                                             name, image_type, position,
-                                            x_dpi, y_dpi, md5)
+                                            x_dpi, y_dpi, digest)
 
             if has_drawing:
                 drawing = sheet.drawing
@@ -1221,7 +1278,7 @@ class Workbook(xmlwriter.XMLwriter):
             # Read the image data from the user supplied byte stream.
             data = image_data.getvalue()
 
-        md5 = hashlib.md5(data).hexdigest()
+        digest = hashlib.sha256(data).hexdigest()
 
         # Get the image filename without the path.
         image_name = os.path.basename(filename)
@@ -1232,10 +1289,12 @@ class Workbook(xmlwriter.XMLwriter):
         marker3 = unpack('2s', data[:2])[0]
         marker4 = unpack('<L', data[:4])[0]
         marker5 = (unpack('4s', data[40:44]))[0]
+        marker6 = unpack('4s', data[:4])[0]
 
         png_marker = b'PNG'
         bmp_marker = b'BM'
         emf_marker = b' EMF'
+        gif_marker = b'GIF8'
 
         if marker1 == png_marker:
             self.image_types['png'] = True
@@ -1257,6 +1316,10 @@ class Workbook(xmlwriter.XMLwriter):
             self.image_types['emf'] = True
             (image_type, width, height, x_dpi, y_dpi) = self._process_emf(data)
 
+        elif marker6 == gif_marker:
+            self.image_types['gif'] = True
+            (image_type, width, height, x_dpi, y_dpi) = self._process_gif(data)
+
         else:
             raise UnsupportedImageFormat(
                 "%s: Unknown or unsupported image file format." % filename)
@@ -1275,7 +1338,7 @@ class Workbook(xmlwriter.XMLwriter):
         if y_dpi == 0:
             y_dpi = 96
 
-        return image_type, width, height, image_name, x_dpi, y_dpi, md5
+        return image_type, width, height, image_name, x_dpi, y_dpi, digest
 
     def _process_png(self, data):
         # Extract width and height information from a PNG file.
@@ -1287,25 +1350,20 @@ class Workbook(xmlwriter.XMLwriter):
         x_dpi = 96
         y_dpi = 96
 
-        # Look for numbers rather than strings for Python 2.6/3 compatibility.
-        marker_ihdr = 0x49484452  # IHDR
-        marker_phys = 0x70485973  # pHYs
-        marker_iend = 0X49454E44  # IEND
-
         # Search through the image data to read the height and width in the
         # IHDR element. Also read the DPI in the pHYs element.
         while not end_marker and offset < data_length:
 
             length = unpack('>I', data[offset + 0:offset + 4])[0]
-            marker = unpack('>I', data[offset + 4:offset + 8])[0]
+            marker = unpack('4s', data[offset + 4:offset + 8])[0]
 
             # Read the image dimensions.
-            if marker == marker_ihdr:
+            if marker == b'IHDR':
                 width = unpack('>I', data[offset + 8:offset + 12])[0]
                 height = unpack('>I', data[offset + 12:offset + 16])[0]
 
             # Read the image DPI.
-            if marker == marker_phys:
+            if marker == b'pHYs':
                 x_density = unpack('>I', data[offset + 8:offset + 12])[0]
                 y_density = unpack('>I', data[offset + 12:offset + 16])[0]
                 units = unpack('b', data[offset + 16:offset + 17])[0]
@@ -1314,7 +1372,7 @@ class Workbook(xmlwriter.XMLwriter):
                     x_dpi = x_density * 0.0254
                     y_dpi = y_density * 0.0254
 
-            if marker == marker_iend:
+            if marker == b'IEND':
                 end_marker = True
                 continue
 
@@ -1375,6 +1433,16 @@ class Workbook(xmlwriter.XMLwriter):
 
         return 'jpeg', width, height, x_dpi, y_dpi
 
+    def _process_gif(self, data):
+        # Extract width and height information from a GIF file.
+        x_dpi = 96
+        y_dpi = 96
+
+        width = unpack('<h', data[6:8])[0]
+        height = unpack('<h', data[8:10])[0]
+
+        return 'gif', width, height, x_dpi, y_dpi
+
     def _process_bmp(self, data):
         # Extract width and height information from a BMP file.
         width = unpack('<L', data[18:22])[0]
@@ -1530,6 +1598,12 @@ class Workbook(xmlwriter.XMLwriter):
             sheet._prepare_tables(table_id + 1, seen)
             table_id += table_count
 
+    def _prepare_metadata(self):
+        # Set the metadata rel link.
+        for sheet in self.worksheets():
+            if sheet.has_dynamic_arrays:
+                self.has_metadata = True
+
     def _add_chart_data(self):
         # Add "cached" data to charts to provide the numCache and strCache
         # data for series and title/axis ranges.
@@ -1584,8 +1658,7 @@ class Workbook(xmlwriter.XMLwriter):
                 # in a chart series formula.
                 if sheetname not in worksheets:
                     warn("Unknown worksheet reference '%s' in range "
-                         "'%s' passed to add_series()"
-                         % (force_unicode(sheetname), force_unicode(c_range)))
+                         "'%s' passed to add_series()" % (sheetname, c_range))
                     chart.formula_data[r_id] = []
                     seen_ranges[c_range] = []
                     continue
@@ -1690,6 +1763,15 @@ class Workbook(xmlwriter.XMLwriter):
 
         self._xml_empty_tag('fileVersion', attributes)
 
+    def _write_file_sharing(self):
+        # Write the <fileSharing> element.
+        if self.read_only == 0:
+            return
+
+        attributes = [('readOnlyRecommended', 1)]
+
+        self._xml_empty_tag('fileSharing', attributes)
+
     def _write_workbook_pr(self):
         # Write <workbookPr> element.
         default_theme_version = 124226
